# Copyright (C) 2001-2006 Python Software Foundation
# Author: Barry Warsaw
# Contact: email-sig@python.org

"""Class representing text/* type MIME documents."""
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

__all__ = ['MIMEText']

from future.backports.email.encoders import encode_7or8bit
from future.backports.email.mime.nonmultipart import MIMENonMultipart


class MIMEText(MIMENonMultipart):
    """Class for generating text/* type MIME documents."""

    def __init__(self, _text, _subtype='plain', _charset=None):
        """Create a text/* type MIME document.

        _text is the string for this message object.

        _subtype is the MIME sub content type, defaulting to "plain".

        _charset is the character set parameter added to the Content-Type
        header.  This defaults to "us-ascii".  Note that as a side-effect, the
        Content-Transfer-Encoding header will also be set.
        """

        # If no _charset was specified, check to see if there are non-ascii
        # characters present. If not, use 'us-ascii', otherwise use utf-8.
        # XXX: This can be removed once #7304 is fixed.
        if _charset is None:
            try:
                _text.encode('us-ascii')
                _charset = 'us-ascii'
            except UnicodeEncodeError:
                _charset = 'utf-8'

        MIMENonMultipart.__init__(self, 'text', _subtype,
                                  **{'charset': _charset})

        self.set_payload(_text, _charset)



import hashlib

from ._compat import text_type
from ._json import json
from .encoding import want_bytes
from .exc import BadPayload
from .exc import BadSignature
from .signer import Signer


def is_text_serializer(serializer):
    """Checks whether a serializer generates text or binary."""
    return isinstance(serializer.dumps({}), text_type)


class Serializer(object):
    """This class provides a serialization interface on top of the
    signer. It provides a similar API to json/pickle and other modules
    but is structured differently internally. If you want to change the
    underlying implementation for parsing and loading you have to
    override the :meth:`load_payload` and :meth:`dump_payload`
    functions.

    This implementation uses simplejson if available for dumping and
    loading and will fall back to the standard library's json module if
    it's not available.

    You do not need to subclass this class in order to switch out or
    customize the :class:`.Signer`. You can instead pass a different
    class to the constructor as well as keyword arguments as a dict that
    should be forwarded.

    .. code-block:: python

        s = Serializer(signer_kwargs={'key_derivation': 'hmac'})

    You may want to upgrade the signing parameters without invalidating
    existing signatures that are in use. Fallback signatures can be
    given that will be tried if unsigning with the current signer fails.

    Fallback signers can be defined by providing a list of
    ``fallback_signers``. Each item can be one of the following: a
    signer class (which is instantiated with ``signer_kwargs``,
    ``salt``, and ``secret_key``), a tuple
    ``(signer_class, signer_kwargs)``, or a dict of ``signer_kwargs``.

    For example, this is a serializer that signs using SHA-512, but will
    unsign using either SHA-512 or SHA1:

    .. code-block:: python

        s = Serializer(
            signer_kwargs={"digest_method": hashlib.sha512},
            fallback_signers=[{"digest_method": hashlib.sha1}]
        )

    .. versionchanged:: 0.14:
        The ``signer`` and ``signer_kwargs`` parameters were added to
        the constructor.

    .. versionchanged:: 1.1.0:
        Added support for ``fallback_signers`` and configured a default
        SHA-512 fallback. This fallback is for users who used the yanked
        1.0.0 release which defaulted to SHA-512.
    """

    #: If a serializer module or class is not passed to the constructor
    #: this one is picked up. This currently defaults to :mod:`json`.
    default_serializer = json

    #: The default :class:`Signer` class that is being used by this
    #: serializer.
    #:
    #: .. versionadded:: 0.14
    default_signer = Signer

    #: The default fallback signers.
    default_fallback_signers = [{"digest_method": hashlib.sha512}]

    def __init__(
        self,
        secret_key,
        salt=b"itsdangerous",
        serializer=None,
        serializer_kwargs=None,
        signer=None,
        signer_kwargs=None,
        fallback_signers=None,
    ):
        self.secret_key = want_bytes(secret_key)
        self.salt = want_bytes(salt)
        if serializer is None:
            serializer = self.default_serializer
        self.serializer = serializer
        self.is_text_serializer = is_text_serializer(serializer)
        if signer is None:
            signer = self.default_signer
        self.signer = signer
        self.signer_kwargs = signer_kwargs or {}
        if fallback_signers is None:
            fallback_signers = list(self.default_fallback_signers or ())
        self.fallback_signers = fallback_signers
        self.serializer_kwargs = serializer_kwargs or {}

    def load_payload(self, payload, serializer=None):
        """Loads the encoded object. This function raises
        :class:`.BadPayload` if the payload is not valid. The
        ``serializer`` parameter can be used to override the serializer
        stored on the class. The encoded ``payload`` should always be
        bytes.
        """
        if serializer is None:
            serializer = self.serializer
            is_text = self.is_text_serializer
        else:
            is_text = is_text_serializer(serializer)
        try:
            if is_text:
                payload = payload.decode("utf-8")
            return serializer.loads(payload)
        except Exception as e:
            raise BadPayload(
                "Could not load the payload because an exception"
                " occurred on unserializing the data.",
                original_error=e,
            )

    def dump_payload(self, obj):
        """Dumps the encoded object. The return value is always bytes.
        If the internal serializer returns text, the value will be
        encoded as UTF-8.
        """
        return want_bytes(self.serializer.dumps(obj, **self.serializer_kwargs))

    def make_signer(self, salt=None):
        """Creates a new instance of the signer to be used. The default
        implementation uses the :class:`.Signer` base class.
        """
        if salt is None:
            salt = self.salt
        return self.signer(self.secret_key, salt=salt, **self.signer_kwargs)

    def iter_unsigners(self, salt=None):
        """Iterates over all signers to be tried for unsigning. Starts
        with the configured signer, then constructs each signer
        specified in ``fallback_signers``.
        """
        if salt is None:
            salt = self.salt
        yield self.make_signer(salt)
        for fallback in self.fallback_signers:
            if type(fallback) is dict:
                kwargs = fallback
                fallback = self.signer
            elif type(fallback) is tuple:
                fallback, kwargs = fallback
            else:
                kwargs = self.signer_kwargs
            yield fallback(self.secret_key, salt=salt, **kwargs)

    def dumps(self, obj, salt=None):
        """Returns a signed string serialized with the internal
        serializer. The return value can be either a byte or unicode
        string depending on the format of the internal serializer.
        """
        payload = want_bytes(self.dump_payload(obj))
        rv = self.make_signer(salt).sign(payload)
        if self.is_text_serializer:
            rv = rv.decode("utf-8")
        return rv

    def dump(self, obj, f, salt=None):
        """Like :meth:`dumps` but dumps into a file. The file handle has
        to be compatible with what the internal serializer expects.
        """
        f.write(self.dumps(obj, salt))

    def loads(self, s, salt=None):
        """Reverse of :meth:`dumps`. Raises :exc:`.BadSignature` if the
        signature validation fails.
        """
        s = want_bytes(s)
        last_exception = None
        for signer in self.iter_unsigners(salt):
            try:
                return self.load_payload(signer.unsign(s))
            except BadSignature as err:
                last_exception = err
        raise last_exception

    def load(self, f, salt=None):
        """Like :meth:`loads` but loads from a file."""
        return self.loads(f.read(), salt)

    def loads_unsafe(self, s, salt=None):
        """Like :meth:`loads` but without verifying the signature. This
        is potentially very dangerous to use depending on how your
        serializer works. The return value is ``(signature_valid,
        payload)`` instead of just the payload. The first item will be a
        boolean that indicates if the signature is valid. This function
        never fails.

        Use it for debugging only and if you know that your serializer
        module is not exploitable (for example, do not use it with a
        pickle serializer).

        .. versionadded:: 0.15
        """
        return self._loads_unsafe_impl(s, salt)

    def _loads_unsafe_impl(self, s, salt, load_kwargs=None, load_payload_kwargs=None):
        """Low level helper function to implement :meth:`loads_unsafe`
        in serializer subclasses.
        """
        try:
            return True, self.loads(s, salt=salt, **(load_kwargs or {}))
        except BadSignature as e:
            if e.payload is None:
                return False, None
            try:
                return (
                    False,
                    self.load_payload(e.payload, **(load_payload_kwargs or {})),
                )
            except BadPayload:
                return False, None

    def load_unsafe(self, f, *args, **kwargs):
        """Like :meth:`loads_unsafe` but loads from a file.

        .. versionadded:: 0.15
        """
        return self.loads_unsafe(f.read(), *args, **kwargs)



from old_helper import convert_from_localtime_to_utc
from typing import List

from commons.report import Generator
from domain import Clock, OperatorTip, TipRepository, OperatorNameRetriever
from utils import get_store_id

from .dto import \
    TipReportDto, \
    TipReportFilterDto, \
    TipReportHeaderDto, \
    TipReportBodyDto, \
    TipReportDetailLineDto


class TipReportGenerator(Generator):
    def __init__(self, filter, operator_name_retriever, clock, tip_repository):
        # type: (TipReportFilterDto, OperatorNameRetriever, Clock, TipRepository) -> None
        self.filter = filter
        self.clock = clock
        self.tip_repository = tip_repository
        self.operator_name_retriever = operator_name_retriever

    def generate_data(self):
        # type: () -> TipReportDto
        header = self._generate_header()
        body = self._generate_body()
        return TipReportDto(header, body)

    def _generate_header(self):
        # type: () -> TipReportHeaderDto
        return TipReportHeaderDto(
            self.filter.start_date,
            self.filter.end_date,
            self.clock.get_current_datetime(),
            get_store_id()
        )

    def _generate_body(self):
        # type: () -> TipReportBodyDto

        tip_report_body_dto = TipReportBodyDto()

        start_date = self.filter.start_date.replace(hour=00, minute=00, second=00)
        start_date = convert_from_localtime_to_utc(start_date)

        end_date = self.filter.end_date.replace(hour=23, minute=59, second=59)
        end_date = convert_from_localtime_to_utc(end_date)

        operator_tip_list = self.tip_repository.get_tips_per_operator(
            start_date,
            end_date,
            self.filter.operator_id)

        operators = set(map(lambda x: x.operator, operator_tip_list))
        grouped_operators_with_tip_records = [[y for y in operator_tip_list if y.operator == x] for x in operators]

        for operator_with_tip_records in grouped_operators_with_tip_records:
            tip_report_detail_line = self._convert_operator_tip_list_to_tip_report_detail_line_dto(
                operator_with_tip_records)
            tip_report_body_dto.add_detail_line(tip_report_detail_line)

        return tip_report_body_dto

    def _convert_operator_tip_list_to_tip_report_detail_line_dto(self, operator_tip_list):
        # type: (List[OperatorTip]) -> TipReportDetailLineDto

        operator_name = self.operator_name_retriever.get_operator_name(operator_tip_list[0].operator)

        tables_count = len(set(map(lambda o_tip: o_tip.service_id, operator_tip_list)))
        order_count = len(operator_tip_list)
        total_order_amount = sum([order_amount.default_unit_price for order_amount in operator_tip_list])
        total_tip_amount = sum(
            [(order_amount.default_unit_price * order_amount.tip_portion) for order_amount in operator_tip_list])
        return TipReportDetailLineDto(
            operator_name,
            tables_count,
            order_count,
            round(total_order_amount, 2),
            round(total_tip_amount, 2)
        )



# encoding: utf-8
import time
from cStringIO import StringIO
from datetime import datetime

from old_helper import convert_from_utf_to_localtime
from utils import (
    report_default_header,
    DATE_TIME_FMT,
    center,
    add_key_value_line,
    empty_line,
    add_table_line,
    get_store_id,
    single_line_separator
)


class OpenedTablesReport(object):
    def __init__(self, table_repository, order_repository, pos_id, operator_name_retriever):
        self.table_repository = table_repository
        self.order_repository = order_repository
        self.pos_id = pos_id
        self.operator_name_retriever = operator_name_retriever
        self.store_id = get_store_id()

        self.columns_height = [18, 20]
        self.character_to_add = [".:", " ", " "]
        self.command = "Comanda"

    def generate_opened_tables_report(self, pos_id):
        report = StringIO()
        title = "Relatorio Mesas Abertas"
        current_datetime = time.strftime(DATE_TIME_FMT)
        report.write(report_default_header(title, int(pos_id), None, None, current_datetime, self.store_id))
        self._opened_tables_report_body(report)

        return report.getvalue()

    def _opened_tables_report_body(self, report):
        tables = self.table_repository.get_opened_tables()
        orders = []
        [orders.extend(x.orders) for x in tables.values() if x.orders != [None]]
        orders_amount = self.order_repository.get_orders_amount(orders)

        sectors_details = {}
        for table in sorted(tables.items(), key=lambda x: x[1].name):
            table_id = table[0]
            current_table = table[1]
            is_tab = "TAB" in table_id

            table_context = "Comanda " if is_tab else "Mesa "
            date = datetime.strptime(current_table.opened_date, "%Y-%m-%d %H:%M:%S")
            date = convert_from_utf_to_localtime(date)
            table_name = current_table.name
            table_detail_header = table_context + str(table_name) + " - " + datetime.strftime(date, "%d/%m/%Y %H:%M:%S")

            table_amount = self._get_table_amount(current_table, orders_amount)
            seats = current_table.seats
            columns = ["Total [Assentos]", 'R$ {:>8.2f}'.format(table_amount) + " [" + seats + "]"]

            operator_id = current_table.operator
            operator_name = self.operator_name_retriever.get_operator_name(operator_id)
            table_info = dict(header=table_detail_header, operator=operator_name, columns=columns)

            sector = self._get_sector_name(current_table, is_tab)
            if sector in sectors_details:
                sectors_details[sector].append(table_info)
            else:
                sectors_details[sector] = [table_info]

        sectors_keys = [s for s in sectors_details.keys() if s != self.command]
        sectors_keys.sort()
        if self.command in sectors_details.keys():
            sectors_keys.append(self.command)

        for sector in sectors_keys:
            report.write(center(sector.upper()))
            report.write(empty_line())
            for table_info in sectors_details[sector]:
                report.write(center(table_info["header"]))
                report.write(add_key_value_line("Operador........: ", table_info["operator"]))
                report.write(add_table_line(
                    table_info["columns"],
                    self.columns_height,
                    character_to_add=self.character_to_add)
                )
                report.write(empty_line())

            report.write(single_line_separator())

    @staticmethod
    def _get_sector_name(current_table, is_tab):
        sector = current_table.sector
        if sector:
            return sector

        if is_tab:
            return 'COMANDAS'

        return 'SEM SETOR'

    @staticmethod
    def _get_table_amount(current_table, orders_amount):
        table_amount = 0
        for order_id in current_table.orders:
            if order_id is None or order_id not in orders_amount:
                continue

            table_amount += orders_amount[order_id]["amount"]
            if current_table.status == 'TOTALED':
                table_amount += orders_amount[order_id]["tip"]

        return table_amount



import json
import time

from datetime import datetime
from logging import Logger
from threading import Condition, Lock
from xml.etree import cElementTree as eTree

import sysactions

from cfgtools import Group
from modelcommon.electronicpayment import TransactionProcessors

from mbcontextmessagehandler import MbContextMessageBus
from msgbus import MBException
from modelcommon.model import SitefResponse
from typing import (
    Any,
    Union,
    Optional,
    Dict,
)
from util import (
    get_transaction_identifier,
    Cache,
)

from ._MobileServiceFinder import MobileServiceFinder
from dialogmanager import DialogManager
from model import (
    CancelPaymentOrderData,
    ConfirmFinishPaymentData,
    FinishPaymentData,
    PaymentResponseMessage,
    TestConnectionData,
)
from paymentprocessor import PaymentProcessor


class MobilePaymentProcessor(PaymentProcessor):

    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
        dialog_manager,     # type: DialogManager
        config,             # type: Group
        logger,             # type: Logger
        cache,              # type: Cache
    ):
        # type: (...) -> None

        super(MobilePaymentProcessor, self).__init__(message_bus, dialog_manager, config, logger, cache)

        sysactions.mbcontext = message_bus.mbcontext
        self.message_bus = message_bus
        self.is_sync = config.find_value("IsMobileSynch", "False") == "True"
        self.sitef_service_finder = MobileServiceFinder()
        self.sync_condition = Condition()
        self.sync_condition_lock = Lock()
        self.has_thread_waiting = False
        self.finalized_status = 0
        self.finalized_data = None
        self.logger = logger
        self.pos_id = config.find_value("PosId")
        self.cache = cache

    def connection_status(
        self,
        test_connection_data    # type: TestConnectionData
    ):
        # type: (...) -> bool

        return True

    def add_payment(self, start_payment_data):
        return PaymentResponseMessage("Mobile Sitef cannot process transactions")

    def cancel_payment_order(
        self,
        cancel_payment_order_data,  # type: CancelPaymentOrderData
        additional_params="",       # type: Optional[str]
    ):
        # type: (...) -> Any[Union[SitefResponse]]

        self.logger.debug("MobileProcessor Cancel payment order - STARTED")

        while self.has_thread_waiting:
            time.sleep(0.2)

        amount = "{:.2f}".format(float(cancel_payment_order_data.tender_amount)).replace(",", "")
        tenders = json.loads(cancel_payment_order_data.tenders)
        tender_is_list = isinstance(tenders, list)
        tenders = [{"detail": tenders[0] if tender_is_list else tenders}]

        data = self._get_json_data(
            cancel_payment_order_data.order_id,
            cancel_payment_order_data.tender_id,
            cancel_payment_order_data.fiscal_date,
            cancel_payment_order_data.fiscal_time,
            tenders,
            is_reversal=True,
            status="0"
        )
        data["tenderType"] = cancel_payment_order_data.tender_type
        data["tenderAmount"] = amount
        data["operator"] = cancel_payment_order_data.operator_id
        data["nsuHost"] = cancel_payment_order_data.nsu_host
        data["cardNumber"] = cancel_payment_order_data.card_number
        data["media"] = cancel_payment_order_data.media
        data = json.dumps(data)
        self.logger.debug("Cancel payment order request: {}", data)

        retry = 0
        while retry < 3:
            try:
                self._publish_event(
                    pos_id=int(cancel_payment_order_data.pos_id),
                    event_type="MOBILE",
                    action_type="MOBILE_REVERSAL_TRANSACTION",
                    event_data=data,
                )
                self._wait_confirmation(
                    seconds_timeout=120
                )
                response = self._get_json_cancel_data(amount=amount)
                if response is not None:
                    self._set_cache_receipt_data(
                        order_id=cancel_payment_order_data.order_id,
                        tender_id=cancel_payment_order_data.tender_id,
                        receipt_customer=response.customer_receipt,
                        receipt_merchant=response.merchant_receipt,
                    )
                    return response
                time.sleep(2)
                break
            except MBException as _:
                self.logger.exception("Error sending mobile cancel transaction")
            finally:
                retry += 1
        self.logger.debug("MobileProcessor Cancel payment order - FINISH")
        return ""

    def cancel_payment(self):
        pass  # Mobile Sitef doesn't handle cancel non-finished transactions

    def finish_payment(
        self,
        finish_payment_data     # type: FinishPaymentData
    ):
        # type: (...) -> bool

        pos_id = finish_payment_data.pos_id
        order_id = finish_payment_data.order_id
        tender_id = finish_payment_data.tender_id
        status = str(finish_payment_data.status)
        fiscal_date = finish_payment_data.fiscal_date
        fiscal_time = finish_payment_data.fiscal_time
        tender_details = finish_payment_data.tender_details

        if self._get_order_state(int(pos_id), order_id, 'VOIDED'):
            return True

        data = self._get_json_data(
            order_id,
            tender_id,
            fiscal_date,
            fiscal_time,
            tender_details,
            is_reversal=False,
            status=status
        )
        data = json.dumps(data)
        retry = 0
        while retry < 3:
            try:
                self._publish_event(
                    int(pos_id),
                    "MOBILE",
                    "MOBILE_CONFIRM_TRANSACTION",
                    data
                )
                break
            except MBException as _:
                self.logger.exception("Error sending mobile confirm transaction")
            finally:
                retry += 1

        if not self.is_sync:
            return True

        return self._wait_confirmation()

    def confirm_finish_payment(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> None

        try:
            self.sitef_service_finder.lock_sitef()
            try_num = 0
            while not self.has_thread_waiting and try_num < 10:
                time.sleep(0.2)
                try_num += 1

            if not self.has_thread_waiting:
                raise Exception("No thread is waiting to finalize")

            with self.sync_condition_lock:
                self.finalized_status = confirm_finish_payment_data.status
                self.finalized_data = confirm_finish_payment_data.data
                with self.sync_condition:
                    self.sync_condition.notify_all()
        except Exception as ex:
            self.logger.exception("Error cancel payment order: {}".format(ex))
        finally:
            self.sitef_service_finder.release_sitef()

    def _publish_event(
        self,
        pos_id,         # type: int
        event_type,     # type: str
        action_type,    # type: str
        event_data      # type: str
    ):
        # type: (...) -> None

        event_subject = "POS{}".format(pos_id)
        event_data = str(event_data)
        event_xml = """<Event subject="{0}" type="{1}">
                               <{2}>
                                   {3}
                               </{2}>
                          </Event>""".format(event_subject, event_type, action_type, event_data)

        self.message_bus.mbcontext.MB_EasyEvtSend(
            subject=event_subject,
            type=event_type,
            xml=event_xml
        )

    @staticmethod
    def _get_json_data(
        order_id,           # type: str
        tender_id,          # type: str
        fiscal_date,        # type: str
        fiscal_time,        # type: str
        tenders,            # type: str
        is_reversal=False,  # type: bool
        status=None         # type: str
    ):
        # type: (...) -> Dict

        if is_reversal:
            fiscal_date = datetime.strptime(fiscal_date, '%Y%m%d').strftime('%d%m%Y')

        transaction_identifier = get_transaction_identifier(order_id, tender_id)
        transaction_data = {
            "timeStamp": time.time(),
            "transactionIdentifier": transaction_identifier,
            "fiscalDate": fiscal_date,
            "fiscalHour": fiscal_time,
            "fiscalTime": fiscal_time,
            "tenders": tenders,
            "isReversal": is_reversal
        }

        if status:
            transaction_data["confirmTransaction"] = str(status)

        return transaction_data

    @staticmethod
    def _get_order_state(
        pos_id,     # type: int
        order_id,   # type: str
        state,      # type: str
    ):
        # type: (...) -> Optional[str]

        from sysactions import get_model, get_posot

        order = get_posot(get_model(pos_id)).orderPicture(orderid=order_id)
        order = eTree.XML(order).find("Order")
        return order.find("StateHistory/State[@state='{}']".format(state)) is not None

    def _wait_confirmation(
        self,
        seconds_timeout=540  # type: int
    ):
        # type: (...) -> int

        self.sync_condition_lock.acquire()
        acquired = True
        try:
            if not self.sync_condition.acquire(False):
                raise Exception("Another thread has the finalize lock")
            try:
                self.has_thread_waiting = True
                self.sync_condition_lock.release()
                acquired = False
                if self.sync_condition.wait(seconds_timeout):
                    raise Exception("Timeout waiting for payment conclusion")
                self.has_thread_waiting = False
                return 0
            finally:
                self.sync_condition.release()
        finally:
            if acquired:
                self.sync_condition_lock.release()

    def get_name(
        self,
    ):
        # type: (...) -> str

        self.register_eft_processor(
            pos_id=self.pos_id,
            processor_name=TransactionProcessors.MOBILE,
        )

        return TransactionProcessors.MOBILE

    def get_document(
        self,
        pos_id      # type: str
    ):
        # type: (...) -> str

        pass

    def stop(
        self,
    ):
        # type: (...) -> None

        pass

    def get_last_receipt(
        self,
        eft_processor,      # type: str
    ):
        # type: (...) -> Optional[str]

        if str(eft_processor.upper()) not in [TransactionProcessors.SITEF, TransactionProcessors.CAPPTA]:
            self._publish_event(
                pos_id=int(self.pos_id),
                event_type="MOBILE",
                action_type="MOBILE_PRINT_LAST_RECEIPT",
                event_data="",
            )

        return self.cache.get_last_receipt(eft_processor=eft_processor)

    def _get_json_cancel_data(
        self,       # type: MobilePaymentProcessor
        amount,     # type: str
    ):
        # type: (...) -> Optional[SitefResponse]

        if self.finalized_status == 0 or self.finalized_data in [None, '', 'None']:
            return None

        self.logger.debug("Cancel payment order response: {}", self.finalized_data)
        data_cancel = json.loads(self.finalized_data)
        response = SitefResponse()
        response.transaction_identifier = data_cancel.get("TransactionIdentifier", "")
        response.nsu = data_cancel.get("Nsu", "")
        response.nsu_host = data_cancel.get("NsuHost", "")
        response.nsu_origin = data_cancel.get("AuthCode", "")
        response.transaction_value = amount
        response.customer_receipt = data_cancel.get("ReceiptCustomer", "")
        response.merchant_receipt = data_cancel.get("ReceiptMerchant", "")

        return response

    def _set_cache_receipt_data(
        self,               # type: MobilePaymentProcessor
        order_id,           # type: str
        tender_id,          # type: str
        receipt_customer,   # type: str
        receipt_merchant,   # type: str
    ):
        # type: (...) -> None

        receipt_data = {
            "order_id": order_id,
            "tender_id": tender_id,
            "receipt_customer": receipt_customer,
            "receipt_merchant": receipt_merchant,
        }

        receipt_data = json.dumps(receipt_data)
        self.cache.set_last_receipt(
            receipt_data=receipt_data,
            eft_processor=TransactionProcessors.SITEF,
        )



from xml.etree import cElementTree as eTree

from application.model import (
    BenefitAppliers,
    OperationDescription,
)
from application.model.configuration import Configurations
from application.model.repository.benefit_applier import DefaultBenefitApplierRepository
from application.repository.pos import DBRepository
from typing import (
    Optional,
    Dict,
)

from application.services import TableService


class BurnBenefitInteractor(object):
    def __init__(
        self,                               # type: BurnBenefitInteractor
        configs,                            # type: Configurations
        db_repository,                      # type: DBRepository
        benefit_appliers_repositories,      # type: Dict[BenefitAppliers: DefaultBenefitApplierRepository]
        table_service,                      # type: TableService
    ):
        # type: (...) -> None

        self.configs = configs
        self.logger = self.configs.logger
        self.db_repository = db_repository
        self.benefit_appliers_repositories = benefit_appliers_repositories
        self.table_service = table_service

    def burn_benefit(
        self,                   # type: BurnBenefitInteractor
        benefit_id,             # type: str
        originator_pos_id,      # type: int
        benefit_applier,        # type: str
        order_picture,          # type: eTree
        table_picture=None,     # type: Optional[eTree]
        retry_quantity=0        # type: Optional[int]
    ):
        # type: (...) -> None

        success = False
        applier_name = benefit_applier.split("|")[0]
        try:
            message = "Burning benefit. BenefitId: {}; OriginatorPosId: {}; BenefitApplier: {}; Try: {}"
            self.logger.info(message.format(benefit_id, originator_pos_id, applier_name, retry_quantity))
            if benefit_applier not in [None, 'None', '']:
                if not table_picture:
                    table_picture = self.table_service.get_table_picture(
                        pos_id=originator_pos_id,
                        order_picture=order_picture,
                    )

                self.benefit_appliers_repositories[applier_name].burn_benefit(
                    benefit_id=benefit_id,
                    order_picture=order_picture,
                    table_picture=table_picture,
                )
            success = True
        except Exception as _:
            self.logger.exception("Error burning benefit: {}".format(benefit_id))
        finally:
            self.db_repository.manage_operation_status(
                success,
                benefit_id,
                OperationDescription.BURNED,
                applier_name,
                retry_quantity,
                order_picture=order_picture,
                table_picture=table_picture,
            )



import sys
from logging import Logger

from cfgtools import Group
from dialogmanager import DialogManager
from messagebus import MessageBus
from model import TestConnectionData
from modelcommon.customexception import SitefFinishPaymentRefundCanceled
from typing import (
    Optional,
    Dict,
    List,
)

import paymentprocessor
from util import Cache


class PaymentError(RuntimeError):
    def __repr__(self):
        return self.message


class PaymentNotMapped(PaymentError):
    def __init__(self, tender_id):
        self.message = "Payment {} not mapped".format(tender_id)

    def __repr__(self):
        return self.message


class ConfigurationNotFound(PaymentError):
    def __init__(self, configuration):
        self.message = "Configuration {} not found".format(configuration)


class PaymentNotFound(PaymentError):
    def __init__(self, payment_type):
        self.message = "Payment type {} not found".format(payment_type)


def _get_payment_class_reference(group_name):
    class_ref = getattr(sys.modules["paymentprocessor"], "{}PaymentProcessor".format(group_name), None)
    if class_ref is None:
        raise PaymentNotFound(group_name)
    return class_ref


class PaymentRouter(object):
    def __init__(
        self,
        message_bus,        # type: MessageBus
        dialog_manager,     # type: DialogManager
        config,             # type: Group
        logger,             # type: Logger
        cache,              # type: Cache
    ):
        self.message_bus = message_bus
        self.dialog_manager = dialog_manager
        self._config = config
        self._logger = logger
        self.cache = cache

        self._electronic_types = {}
        self._current_payment_processor = None  # type: Optional[paymentprocessor.PaymentProcessor]
        self._payment_route = {}                # type: Dict[int, paymentprocessor.PaymentProcessor]
        self._payment_processors = []           # type: List[paymentprocessor.PaymentProcessor]
        self._ask_via_pin_pad_routes = {}       # type: Dict[str, paymentprocessor.PaymentProcessor]
        self._fill_payment_routes()

    def _fill_payment_routes(self):
        try:
            payment_types = self._config.find_group("PaymentTypes")
            if not payment_types:
                raise ConfigurationNotFound("PaymentTypes")
            ask_document_processor = self._config.find_value("AskDocumentProcessor")

            for group in payment_types.groups:
                class_ref = _get_payment_class_reference(group_name=group.name)
                class_instance = class_ref(
                    message_bus=self.message_bus,
                    dialog_manager=self.dialog_manager,
                    config=group,
                    logger=self._logger,
                    cache=self.cache,
                )
                self._payment_processors.append(class_instance)
                for electronic_type in group.find_values("ElectronicTypes") or []:
                    self._electronic_types[int(electronic_type)] = True
                    self._payment_route[int(electronic_type)] = class_instance

                if ask_document_processor == group.name:
                    self._ask_via_pin_pad_routes["document"] = class_instance
        except Exception as _:
            self._logger.exception("Exception creating payment routes")
            raise

    def get_payment_processor(self, electronic_type=None):
        # type: (Optional[int]) -> paymentprocessor.PaymentProcessor

        if electronic_type is None:
            return self._current_payment_processor

        if electronic_type not in self._payment_route:
            raise PaymentNotMapped(electronic_type)

        self._current_payment_processor = self._payment_route[electronic_type]
        return self._current_payment_processor

    def get_ask_via_pin_pad_processor(
        self,
        route   # type: str
    ):
        # type: (...) -> paymentprocessor.PaymentProcessor

        if route not in self._ask_via_pin_pad_routes:
            raise PaymentNotMapped(route)

        return self._ask_via_pin_pad_routes[route]

    def get_connection_status(
        self,
        test_connection_data  # type: TestConnectionData
    ):
        # type: (...) -> bool

        validated_ips = []
        for payment_processor in self._payment_route.itervalues():
            ip_key = [x.values[0] for x in payment_processor.config.keys if x.name == "IP"]
            if len(ip_key) > 0:
                ip = ip_key[0]
                if ip in validated_ips:
                    continue

                validated_ips.append(ip)

            if payment_processor.connection_status(test_connection_data):
                return True

        return False

    def get_electronic_types(
        self
    ):
        # type: (...) -> Dict[int, bool]

        return self._electronic_types

    def get_name(
        self,
    ):
        # type: (...) -> str

        for payment_processor in self._payment_route.itervalues():
            return payment_processor.get_name()

    def get_last_receipt(
        self,
        eft_processor,     # type: str
    ):
        # type: (...) -> Optional[str]

        for payment_processor in self._payment_route.itervalues():
            return payment_processor.get_last_receipt(eft_processor=eft_processor)

    def stop_processors(self):
        for instance in self._payment_processors:
            instance.stop()

    def finish_all_payments(self, data):
        for processor in self._payment_processors:
            if not processor.finish_payment(data):
                raise SitefFinishPaymentRefundCanceled()

    def confirm_finish_payment(self, data):
        for processor in self._payment_processors:
            processor.confirm_finish_payment(data)



# BER encoder
from pyasn1.type import base, tag, univ, char, useful
from pyasn1.codec.ber import eoo
from pyasn1.compat.octets import int2oct, oct2int, ints2octs, null, str2octs
from pyasn1 import debug, error

class Error(Exception): pass

class AbstractItemEncoder:
    supportIndefLenMode = 1
    def encodeTag(self, t, isConstructed):
        tagClass, tagFormat, tagId = t.asTuple()  # this is a hotspot
        v = tagClass | tagFormat
        if isConstructed:
            v = v|tag.tagFormatConstructed
        if tagId < 31:
            return int2oct(v|tagId)
        else:
            s = int2oct(tagId&0x7f)
            tagId = tagId >> 7
            while tagId:
                s = int2oct(0x80|(tagId&0x7f)) + s
                tagId = tagId >> 7
            return int2oct(v|0x1F) + s

    def encodeLength(self, length, defMode):
        if not defMode and self.supportIndefLenMode:
            return int2oct(0x80)
        if length < 0x80:
            return int2oct(length)
        else:
            substrate = null
            while length:
                substrate = int2oct(length&0xff) + substrate
                length = length >> 8
            substrateLen = len(substrate)
            if substrateLen > 126:
                raise Error('Length octets overflow (%d)' % substrateLen)
            return int2oct(0x80 | substrateLen) + substrate

    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        raise Error('Not implemented')

    def _encodeEndOfOctets(self, encodeFun, defMode):
        if defMode or not self.supportIndefLenMode:
            return null
        else:
            return encodeFun(eoo.endOfOctets, defMode)
        
    def encode(self, encodeFun, value, defMode, maxChunkSize):
        substrate, isConstructed = self.encodeValue(
            encodeFun, value, defMode, maxChunkSize
            )
        tagSet = value.getTagSet()
        if tagSet:
            if not isConstructed:  # primitive form implies definite mode
                defMode = 1
            return self.encodeTag(
                tagSet[-1], isConstructed
                ) + self.encodeLength(
                len(substrate), defMode
                ) + substrate + self._encodeEndOfOctets(encodeFun, defMode)
        else:
            return substrate  # untagged value

class EndOfOctetsEncoder(AbstractItemEncoder):
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        return null, 0

class ExplicitlyTaggedItemEncoder(AbstractItemEncoder):
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        if isinstance(value, base.AbstractConstructedAsn1Item):
            value = value.clone(tagSet=value.getTagSet()[:-1],
                                cloneValueFlag=1)
        else:
            value = value.clone(tagSet=value.getTagSet()[:-1])
        return encodeFun(value, defMode, maxChunkSize), 1

explicitlyTaggedItemEncoder = ExplicitlyTaggedItemEncoder()

class BooleanEncoder(AbstractItemEncoder):
    supportIndefLenMode = 0
    _true = ints2octs((1,))
    _false = ints2octs((0,))
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        return value and self._true or self._false, 0

class IntegerEncoder(AbstractItemEncoder):
    supportIndefLenMode = 0
    supportCompactZero = False
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        if value == 0:  # shortcut for zero value
            if self.supportCompactZero:
                # this seems to be a correct way for encoding zeros
                return null, 0
            else:
                # this seems to be a widespread way for encoding zeros
                return ints2octs((0,)), 0
        octets = []
        value = int(value) # to save on ops on asn1 type
        while 1:
            octets.insert(0, value & 0xff)
            if value == 0 or value == -1:
                break
            value = value >> 8
        if value == 0 and octets[0] & 0x80:
            octets.insert(0, 0)
        while len(octets) > 1 and \
                  (octets[0] == 0 and octets[1] & 0x80 == 0 or \
                   octets[0] == 0xff and octets[1] & 0x80 != 0):
            del octets[0]
        return ints2octs(octets), 0

class BitStringEncoder(AbstractItemEncoder):
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        if not maxChunkSize or len(value) <= maxChunkSize*8:
            out_len = (len(value) + 7) // 8
            out_list = out_len * [0]
            j = 7
            i = -1
            for val in value:
                j += 1
                if j == 8:
                    i += 1
                    j = 0
                out_list[i] = out_list[i] | val << (7-j)
            return int2oct(7-j) + ints2octs(out_list), 0
        else:
            pos = 0; substrate = null
            while 1:
                # count in octets
                v = value.clone(value[pos*8:pos*8+maxChunkSize*8])
                if not v:
                    break
                substrate = substrate + encodeFun(v, defMode, maxChunkSize)
                pos = pos + maxChunkSize
            return substrate, 1

class OctetStringEncoder(AbstractItemEncoder):
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        if not maxChunkSize or len(value) <= maxChunkSize:
            return value.asOctets(), 0
        else:
            pos = 0; substrate = null
            while 1:
                v = value.clone(value[pos:pos+maxChunkSize])
                if not v:
                    break
                substrate = substrate + encodeFun(v, defMode, maxChunkSize)
                pos = pos + maxChunkSize
            return substrate, 1

class NullEncoder(AbstractItemEncoder):
    supportIndefLenMode = 0
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        return null, 0

class ObjectIdentifierEncoder(AbstractItemEncoder):
    supportIndefLenMode = 0
    precomputedValues = {
        (1, 3, 6, 1, 2): (43, 6, 1, 2),        
        (1, 3, 6, 1, 4): (43, 6, 1, 4)
    }
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):    
        oid = value.asTuple()
        if oid[:5] in self.precomputedValues:
            octets = self.precomputedValues[oid[:5]]
            oid = oid[5:]
        else:
            if len(oid) < 2:
                raise error.PyAsn1Error('Short OID %s' % (value,))

            octets = ()

            # Build the first twos
            if oid[0] == 0 and 0 <= oid[1] <= 39:
                oid = (oid[1],) + oid[2:]
            elif oid[0] == 1 and 0 <= oid[1] <= 39:
                oid = (oid[1] + 40,) + oid[2:]
            elif oid[0] == 2:
                oid = (oid[1] + 80,) + oid[2:]
            else:
                raise error.PyAsn1Error(
                    'Impossible initial arcs %s at %s' % (oid[:2], value)
                    )

        # Cycle through subIds
        for subId in oid:
            if subId > -1 and subId < 128:
                # Optimize for the common case
                octets = octets + (subId & 0x7f,)
            elif subId < 0:
                raise error.PyAsn1Error(
                    'Negative OID arc %s at %s' % (subId, value)
                )
            else:
                # Pack large Sub-Object IDs
                res = (subId & 0x7f,)
                subId = subId >> 7
                while subId > 0:
                    res = (0x80 | (subId & 0x7f),) + res
                    subId = subId >> 7 
                # Add packed Sub-Object ID to resulted Object ID
                octets += res

        return ints2octs(octets), 0

class RealEncoder(AbstractItemEncoder):
    supportIndefLenMode = 0
    binEncBase = 2 # set to None to choose encoding base automatically 
    def _dropFloatingPoint(self, m, encbase, e):
        ms, es = 1, 1
        if m < 0:
            ms = -1  # mantissa sign
        if e < 0:
            es = -1  # exponenta sign 
        m *= ms 
        if encbase == 8:
            m = m*2**(abs(e) % 3 * es)
            e = abs(e) // 3 * es
        elif encbase == 16:
            m = m*2**(abs(e) % 4 * es)
            e = abs(e) // 4 * es

        while 1:
            if int(m) != m:
                m *= encbase
                e -= 1
                continue
            break
        return ms, int(m), encbase, e

    def _chooseEncBase(self, value):
        m, b, e = value
        base = [2, 8, 16]
        if value.binEncBase in base:
            return self._dropFloatingPoint(m, value.binEncBase, e)
        elif self.binEncBase in base:
            return self._dropFloatingPoint(m, self.binEncBase, e)
        # auto choosing base 2/8/16 
        mantissa = [m, m, m]
        exponenta = [e, e, e]
        encbase = 2 
        e = float('inf')
        for i in range(3):
            sign, mantissa[i], base[i], exponenta[i] = \
                self._dropFloatingPoint(mantissa[i], base[i], exponenta[i])
            if abs(exponenta[i]) < abs(e) or \
               (abs(exponenta[i]) == abs(e) and mantissa[i] < m):
                e = exponenta[i]
                m = int(mantissa[i])
                encbase = base[i]
        return sign, m, encbase, e

    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        if value.isPlusInfinity():
            return int2oct(0x40), 0
        if value.isMinusInfinity():
            return int2oct(0x41), 0
        m, b, e = value
        if not m:
            return null, 0
        if b == 10:
            return str2octs('\x03%dE%s%d' % (m, e == 0 and '+' or '', e)), 0
        elif b == 2:
            fo = 0x80 # binary encoding
            ms, m, encbase, e = self._chooseEncBase(value)
            if ms < 0: # mantissa sign
                fo = fo | 0x40 # sign bit
            # exponenta & mantissa normalization
            if encbase == 2:
                while m & 0x1 == 0:
                    m >>= 1
                    e += 1
            elif encbase == 8:
                while m & 0x7 == 0:
                    m >>= 3
                    e += 1
                fo |= 0x10
            else: # encbase = 16
                while m & 0xf == 0:
                    m >>= 4
                    e += 1
                fo |= 0x20
            sf = 0 # scale factor
            while m & 0x1 == 0:
                m >>= 1
                sf += 1
            if sf > 3:
                raise error.PyAsn1Error('Scale factor overflow') # bug if raised
            fo |= sf << 2
            eo = null
            if e == 0 or e == -1:
                eo = int2oct(e&0xff)
            else: 
                while e not in (0, -1):
                    eo = int2oct(e&0xff) + eo
                    e >>= 8
                if e == 0 and eo and oct2int(eo[0]) & 0x80:
                    eo = int2oct(0) + eo
                if e == -1 and eo and not (oct2int(eo[0]) & 0x80):
                    eo = int2oct(0xff) + eo
            n = len(eo)
            if n > 0xff:
                raise error.PyAsn1Error('Real exponent overflow')
            if n == 1:
                pass
            elif n == 2:
                fo |= 1
            elif n == 3:
                fo |= 2
            else:
                fo |= 3
                eo = int2oct(n&0xff) + eo
            po = null
            while m:
                po = int2oct(m&0xff) + po
                m >>= 8
            substrate = int2oct(fo) + eo + po
            return substrate, 0
        else:
            raise error.PyAsn1Error('Prohibited Real base %s' % b)

class SequenceEncoder(AbstractItemEncoder):
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        value.setDefaultComponents()
        value.verifySizeSpec()
        substrate = null; idx = len(value)
        while idx > 0:
            idx = idx - 1
            if value[idx] is None:  # Optional component
                continue
            component = value.getDefaultComponentByPosition(idx)
            if component is not None and component == value[idx]:
                continue
            substrate = encodeFun(
                value[idx], defMode, maxChunkSize
                ) + substrate
        return substrate, 1

class SequenceOfEncoder(AbstractItemEncoder):
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        value.verifySizeSpec()
        substrate = null; idx = len(value)
        while idx > 0:
            idx = idx - 1
            substrate = encodeFun(
                value[idx], defMode, maxChunkSize
                ) + substrate
        return substrate, 1

class ChoiceEncoder(AbstractItemEncoder):
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        return encodeFun(value.getComponent(), defMode, maxChunkSize), 1

class AnyEncoder(OctetStringEncoder):
    def encodeValue(self, encodeFun, value, defMode, maxChunkSize):
        return value.asOctets(), defMode == 0

tagMap = {
    eoo.endOfOctets.tagSet: EndOfOctetsEncoder(),
    univ.Boolean.tagSet: BooleanEncoder(),
    univ.Integer.tagSet: IntegerEncoder(),
    univ.BitString.tagSet: BitStringEncoder(),
    univ.OctetString.tagSet: OctetStringEncoder(),
    univ.Null.tagSet: NullEncoder(),
    univ.ObjectIdentifier.tagSet: ObjectIdentifierEncoder(),
    univ.Enumerated.tagSet: IntegerEncoder(),
    univ.Real.tagSet: RealEncoder(),
    # Sequence & Set have same tags as SequenceOf & SetOf
    univ.SequenceOf.tagSet: SequenceOfEncoder(),
    univ.SetOf.tagSet: SequenceOfEncoder(),
    univ.Choice.tagSet: ChoiceEncoder(),
    # character string types
    char.UTF8String.tagSet: OctetStringEncoder(),
    char.NumericString.tagSet: OctetStringEncoder(),
    char.PrintableString.tagSet: OctetStringEncoder(),
    char.TeletexString.tagSet: OctetStringEncoder(),
    char.VideotexString.tagSet: OctetStringEncoder(),
    char.IA5String.tagSet: OctetStringEncoder(),
    char.GraphicString.tagSet: OctetStringEncoder(),
    char.VisibleString.tagSet: OctetStringEncoder(),
    char.GeneralString.tagSet: OctetStringEncoder(),
    char.UniversalString.tagSet: OctetStringEncoder(),
    char.BMPString.tagSet: OctetStringEncoder(),
    # useful types
    useful.ObjectDescriptor.tagSet: OctetStringEncoder(),
    useful.GeneralizedTime.tagSet: OctetStringEncoder(),
    useful.UTCTime.tagSet: OctetStringEncoder()        
    }

# Type-to-codec map for ambiguous ASN.1 types
typeMap = {
    univ.Set.typeId: SequenceEncoder(),
    univ.SetOf.typeId: SequenceOfEncoder(),
    univ.Sequence.typeId: SequenceEncoder(),
    univ.SequenceOf.typeId: SequenceOfEncoder(),
    univ.Choice.typeId: ChoiceEncoder(),
    univ.Any.typeId: AnyEncoder()
    }

class Encoder:
    supportIndefLength = True
    def __init__(self, tagMap, typeMap={}):
        self.__tagMap = tagMap
        self.__typeMap = typeMap

    def __call__(self, value, defMode=True, maxChunkSize=0):
        if not defMode and not self.supportIndefLength:
            raise error.PyAsn1Error('Indefinite length encoding not supported by this codec')
        debug.logger & debug.flagEncoder and debug.logger('encoder called in %sdef mode, chunk size %s for type %s, value:\n%s' % (not defMode and 'in' or '', maxChunkSize, value.prettyPrintType(), value.prettyPrint()))
        tagSet = value.getTagSet()
        if len(tagSet) > 1:
            concreteEncoder = explicitlyTaggedItemEncoder
        else:
            if value.typeId is not None and value.typeId in self.__typeMap:
                concreteEncoder = self.__typeMap[value.typeId]
            elif tagSet in self.__tagMap:
                concreteEncoder = self.__tagMap[tagSet]
            else:
                tagSet = value.baseTagSet
                if tagSet in self.__tagMap:
                    concreteEncoder = self.__tagMap[tagSet]
                else:
                    raise Error('No encoder for %s' % (value,))
        debug.logger & debug.flagEncoder and debug.logger('using value codec %s chosen by %s' % (concreteEncoder.__class__.__name__, tagSet))
        substrate = concreteEncoder.encode(
            self, value, defMode, maxChunkSize
            )
        debug.logger & debug.flagEncoder and debug.logger('built %s octets of substrate: %s\nencoder completed' % (len(substrate), debug.hexdump(substrate)))
        return substrate

encode = Encoder(tagMap, typeMap)



class AppliedBenefitDto(object):
    def __init__(
        self,               # type: AppliedBenefitDto
        response_detail,    # type: str
        benefit_applied,    # type: bool
    ):
        # type: (...) -> None

        self.response_detail = response_detail
        self.benefit_applied = benefit_applied



import json

from xml.etree import cElementTree as eTree

from typing import (
    Optional,
)

from mwhelper import ModelService
from sysactions import (
    get_model,
    get_posot,
    get_podtype,
)


class StaticUnitOfWork(object):
    def __init__(
        self,
        pos_id,         # type: int
        order_picture,  # type: eTree.Element
        model=None,     # type: Optional[eTree.Element]
        table_id=None,  # type: Optional[str]
    ):
        # type: (...) -> None

        self.pos_id = pos_id
        self.table_id = table_id

        self.model = model
        if model is None:
            self.model = get_model(posid=pos_id)

        self.pos_ot = get_posot(model=self.model)
        self.order_picture = order_picture
        self._price_list = None
        self.table_xml = None

    def get_table_info(
        self,
    ):
        # type: (...) -> None

        model = self.model
        if model is None:
            model = get_model(self.pos_id)
        self.table_xml = self._get_table_xml(model)

    def set_order_picture(
        self,
        order_picture,  # type: eTree.Element
    ):
        # type: (...) -> None

        if order_picture is not None:
            self.order_picture = order_picture

    def _get_table_custom_property_info(
        self,
        table_xml,  # type: eTree.Element
    ):
        # type: (...) -> None

        for custom_property in table_xml.findall(".//Properties"):
            service_period = custom_property.get('SERVICE_PERIOD_TABLE', None)
            if service_period:
                self.special_catalog = json.loads(service_period)
                break

    def _get_table_xml(
        self,
        model,  # type: eTree.Element
    ):
        # type: (...) -> Optional[eTree.Element]

        from tablemgrapi import get_pos_ts

        table_id = self.table_id
        if not table_id:
            return

        pos_ts = get_pos_ts(model=model)
        return eTree.XML(pos_ts.getTableInfo(table_id=table_id))

    @property
    def price_list(
        self,
    ):
        # type: (...) -> str

        if self._price_list is None:
            pod_type = get_podtype(model=self.model)
            sale_type = self.order_picture.get("saleTypeDescr") if self.order_picture is not None else None
            model_service = ModelService(pos_id=self.pos_id)
            self._price_list = model_service.get_price_list_from_model(
                pod_type=pod_type,
                sale_type=sale_type,
                model=self.model,
                order_picture=self.order_picture,
            )
        return self._price_list

    @property
    def tip(
        self,
    ):
        # type: (...) -> float

        if not self.table_id:
            return 0

        if self.table_xml is None:
            self.get_table_info()

        return float(self.table_xml.get("tip", 0))

    @property
    def seats(
        self,
    ):
        # type: (...) -> int

        if not self.table_id:
            return 0

        if self.table_xml is None:
            self.get_table_info()

        return int(self.table_xml.get("serviceSeats", 0))

    @property
    def sector(
        self,
    ):
        # type: (...) -> Optional[str]

        if not self.table_id:
            return None

        if self.table_xml is None:
            self.get_table_info()
        return self.table_xml.get("sector", None)



from application.domain import (
    I18nImpl,
    I18n,
)
from application.repository.impl import (
    MwI18nRepository,
)

i18n_cache = {}


def get_i18n(
    pos_id,  # type: int
):
    # type: (...) -> I18n

    global i18n_cache
    if pos_id not in i18n_cache:
        i18n_cache[pos_id] = I18nImpl(MwI18nRepository(pos_id))
    return i18n_cache[pos_id]



import logging

from application.repository.api import APIRepository
from typing import Optional

logger = logging.getLogger(__name__)


class LockBenefitInteractor(object):
    def __init__(
        self,            # type: LockBenefitInteractor
        api_repository,  # type: APIRepository
    ):
        # type: (...) -> None

        self.api_repository = api_repository

    def lock_benefit(
        self,        # type: LockBenefitInteractor
        benefit_id,  # type: str
        pos_id,      # type: int
    ):
        # type: (...) -> None

        logger.info("Locking benefit: {}; PosId: {}".format(benefit_id, pos_id))
        self.api_repository.lock_benefit(
            benefit_id=benefit_id,
            pos_id=pos_id,
        )
        logger.info("Benefit locked")



from typing import List

from ._DeliveryHeaderDto import DeliveryHeaderDto
from ._DeliveryBodyDto import DeliveryBodyDto


class DeliveryGeneratorDto(object):
    def __init__(
        self,
        header,                 # type: DeliveryHeaderDto
        body,                   # type: DeliveryBodyDto
        products,               # type: object
        cfh_products,           # type: List
        show_customer_phone,    # type: bool
    ):
        # type: (...) -> None

        self.header = header
        self.body = body
        self.products = products
        self.cfh_products = cfh_products
        self.show_customer_phone = show_customer_phone



from cashreport.model import Order
from old_helper import BaseRepository
from typing import Dict, List


class TableRepository(BaseRepository):
    def __init__(self, mbcontext):
        super(TableRepository, self).__init__(mbcontext)

    def get_sales_average_info(self, start_date, end_date, operator_id):
        def inner_func(conn):
            cursor = conn.select(self._AverageInfoBetweenPeriodsQuery.format(start_date.strftime("%Y%m%d"),
                                                                             end_date.strftime("%Y%m%d"),
                                                                             operator_id or ""))

            average_info = TableSalesAverageInfo()
            response = cursor.get_row(0)
            if response:
                average_info.total_tables = int(response.get_entry("TotalTables") or 0)
                average_info.total_customers = int(response.get_entry("TotalCustomers") or 0)
                average_info.table_time_average_milliseconds = int(round(float(response.get_entry("TableTimeAverageMilliseconds") or 0)))

            return average_info

        return self.execute_with_connection(inner_func)

    _AverageInfoBetweenPeriodsQuery = """
        SELECT 
            COUNT(TableId) AS TotalTables, 
            SUM(numberofseats) AS TotalCustomers,
            AVG((JULIANDAY(FinishedTS) - JULIANDAY(StartTS)) * 86400000) AS TableTimeAverageMilliseconds
        FROM  tableservice tsvc
        JOIN (
            SELECT 
                serviceid, 
                userid 
            FROM serviceorders 
            WHERE State = 'PAID'
            GROUP BY serviceid 
        ) so ON so.serviceid = tsvc.serviceid
        WHERE strftime('%Y%m%d', tsvc.BusinessPeriod, 'utc') >= '{}' 
        AND strftime('%Y%m%d', tsvc.BusinessPeriod, 'utc') <= '{}'
        AND tsvc.serviceid IN (
            SELECT DISTINCT serviceid 
            FROM ServiceTenders
        ) 
        AND finishedts IS NOT NULL
        AND so.UserId LIKE '%{}%'
    """

    def get_tables_by_sector(self, orders):
        # type: (List[Order]) -> Dict
        def inner_func(conn):
            order_ids = "\',\'".join([str(x.table_id) for x in orders if x.table_id is not None])
            tables_sector = {}
            for row in conn.select(self._GetTablesSector.format(order_ids)):
                tables_sector[row.get_entry(0)] = {"sector": row.get_entry(1), "source_table_id": row.get_entry(2)}

            return tables_sector

        return self.execute_with_connection(inner_func)

    _GetTablesSector = """
select distinct ts.tableid, coalesce(rt.sector, 'Comanda') as Sector, ts2.tableid as SourceTableId
from tableservice ts
left join (select tableid, sector from restauranttable) rt on rt.tableid = ts.tableid
left join (select tableid, serviceid from tableservice) ts2 on ts2.serviceid = ts.sourceserviceid
where ts.tableid in (select tableid from restauranttable)
or ts.tableid in ('{}')"""


class TableSalesAverageInfo(object):
    def __init__(self, total_tables=0, total_customers=0, table_time_average_milliseconds=0):
        self.total_tables = total_tables
        self.total_customers = total_customers
        self.table_time_average_milliseconds = table_time_average_milliseconds



# -*- coding: utf-8 -*-
import base64
import json
import logging
import requests
import datetime

from modelcommon.model import SitefResponse
from typing import (
    Any,
    Union,
    Optional,
)

from model import (
    CancelPaymentOrderData,
    PaymentResponseData,
    PaymentResponseMessage
)
from paymentprocessor import PaymentProcessor

logger = logging.getLogger("ElectronicPayment")


class ConectCarPaymentProcessor(PaymentProcessor):

    def connection_status(self, test_connection_data):
        try:
            init_url = self.config.find_value("URL")
            requests.get(init_url, timeout=7.0)
            return True
        except requests.RequestException:
            return False

    def add_payment(self, start_payment_data):
        try:
            pos_id = start_payment_data.pos_id

            status, placa_json = self.get_placa_conectcar(pos_id)
            if status != 200:
                ret = self.dialog_manager.show_question(pos_id, "$ERROR_CONECTCAR_2",
                                                  buttons="$TRY_AGAIN|$CANCEL")
                if ret == '0':
                    status, placa_json = self.get_placa_conectcar(pos_id)
                else:
                    return PaymentResponseMessage("ConectCar: Pagamento cancelado")

            if status == 200:
                placa = json.loads(placa_json).get('placa')
                without_error, message = self.do_conectcar_payment(placa, start_payment_data.amount,
                                                                       start_payment_data.order_id, pos_id)
                if without_error:
                    return PaymentResponseData(
                        media='',
                        result='0',
                        card_number='4111XXXXXXXXXXXX',
                        auth_code='1302',
                        approved_amount=start_payment_data.amount,
                        id_auth='08496898000142',
                        cnpj_auth="68536379000192",
                        bandeira='4',
                        receipt_merchant=base64.b64encode("Recibo ConectCar"),
                        receipt_customer=base64.b64encode("Recibo ConectCar"),
                        expiration_date='05/2050',
                        nsu='1234',
                        transaction_processor="00051",
                        last_digits="1111",
                        sat_acquirer="012",
                        t_pag="",
                        t_band="01",
                        nsu_host="4321",
                        electronic_type=start_payment_data.electronic_type,
                        transaction_identifier="9876"
                    )
                else:
                    logger.error("Erro no pagamento com ConectCar")
                    return PaymentResponseMessage(message)
            else:
                return PaymentResponseMessage(placa_json)

        except Exception as ex:
            return PaymentResponseMessage("Erro pagamento ConectCar:{}".format(ex.message))

    def cancel_payment_order(
        self,
        cancel_payment_order_data,  # type: CancelPaymentOrderData
        additional_params="",       # type: Optional[str]
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        # TODO: Add implementation
        pass

    def cancel_payment(self):
        return ""

    def finish_payment(self, finish_payment_data):
        if finish_payment_data.status == 0:
            logger.error("Erro ao finalizar pagamento com ConectCar - {}".format(finish_payment_data.pos_id))
        return ""

    def get_name(self):
        return "ConectCarElectronicPayment"

    def get_placa_conectcar(self, pos_id):
        try:
            auth, init_url, timeout = self.get_config_values()
            headers = {'Content-type': 'application/json', 'Authorization': auth}
            logger.info('get_placa_conectcar')
            url = "{}api/consultarleitura/idPdv/{}".format(init_url, pos_id)
            response = requests.get(
                url,
                headers=headers,
                timeout=timeout
            )
            if response.status_code == 200:
                return response.status_code, response.content
            else:
                return response.status_code, "$ERROR_CONECTCAR_3"
        except requests.RequestException:
            return 400, "$ERROR_CONECTCAR_1"

    def get_config_values(self):
        init_url = self.config.find_value("URL")
        auth = self.config.find_value("Auth")
        timeout = int(self.config.find_value("Timeout"))
        return auth, init_url, timeout

    def do_conectcar_payment(self, placa, amount, order_id, pos_id):
        try:
            auth, init_url, timeout = self.get_config_values()
            headers = {'Content-type': 'application/json', 'Authorization': auth}
            logger.info("Pagamento com conectcar: {}".format(order_id))
            float_amount = float(amount)
            payment_request = {
                'idPdv': "{}".format(pos_id),
                'meioPagamento': 'ConectCar',
                'placa': placa,
                'valor': float_amount,
                'data': datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
                'identificador': "{}".format(order_id)
            }
            url = "{}api/pagamento".format(init_url)
            response = requests.post(
                url,
                json=payment_request,
                headers=headers,
                timeout=15
            )
            if response.status_code == 200:
                logger.info("Código transação pagamento com conectcar: {}".format(response.status_code))
                return True, response
            return False, "$ERROR_CONECTCAR_4"
        except Exception:
            logger.error("ConectCar: Pagamento indisponivel. Ofereca outra opcao de pagamento")
            return False, "$ERROR_CONECTCAR_5"



from application.interactor.apply.dto import ApplyContext


class ApplyBenefitRequest(object):
    def __init__(
        self,
        benefit,  # type: str,
        context,  # type: ApplyContext
    ):
        # type: (...) -> None

        self.benefit = benefit
        self.context = context



from json import JSONEncoder

from ._SitefResponse import SitefResponse


class SitefResponseJsonEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, SitefResponse):
            return {
                "Nsu": o.nsu,
                "NsuHost": o.nsu_host,
                "NsuOrigin": o.nsu_origin,
                "TransactionValue": o.transaction_value,
                "ReceiptCustomer": o.customer_receipt,
                "ReceiptMerchant": o.merchant_receipt,
                "TransactionIdentifier": o.transaction_identifier
            }

        return super(SitefResponseJsonEncoder, self).default(o)



from __future__ import absolute_import, division, print_function, unicode_literals

import locale
import sys

sys_encoding = locale.getpreferredencoding()

USING_PYTHON2 = True if sys.version_info < (3, 0) else False

if USING_PYTHON2:
    str = unicode  # noqa
else:
    str = str


def ensure_bytes(x, encoding=sys_encoding):
    if not isinstance(x, bytes):
        x = x.encode(encoding)
    return x


def ensure_str(x, encoding=sys_encoding):
    if not isinstance(x, str):
        x = x.decode(encoding)
    return x



from ._TableReportOrderDto import TableReportOrderDto  # noqa
from typing import List, Dict


class TableReportBodyDto(object):
    def __init__(self):
        self.total_amount = 0.0  # type: float
        self.sub_total_amount = 0.0  # type: float
        self.discount_amount = 0.0  # type: float
        self.detailed_discounts = {}  # type: Dict
        self.detailed_payments = {}  # type: Dict
        self.total_after_discount = 0.0  # type: float
        self.total_tender = 0.0  # type: float
        self.total_per_seat = 0.0  # type: float
        self.orders = []  # type: List[TableReportOrderDto]
        self.service_seats = 0  # type: int
        self.order_id = None  # type: str
        self.tip = None  # type: float
        self.promotional_tenders = 0  # type: float
        self.detailed_tip_excluded = {}  # type: Dict
        self.service_tenders = []  # type: List[dict]
        self.discount_groups = {}  # type: Dict



from abc import ABCMeta, abstractmethod

from messagebus import Event


class EventPublisher(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def new_event(self, event):
        # type: (Event) -> None
        raise NotImplementedError()



from abc import ABCMeta, abstractmethod


class MessageProcessorExecutor(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def execute_processor(self, message_bus, message, event, data):
        raise NotImplementedError()



from ._json import json
from .encoding import base64_decode
from .encoding import base64_encode
from .encoding import want_bytes
from .exc import BadData
from .exc import BadHeader
from .exc import BadPayload
from .exc import BadSignature
from .exc import BadTimeSignature
from .exc import SignatureExpired
from .jws import JSONWebSignatureSerializer
from .jws import TimedJSONWebSignatureSerializer
from .serializer import Serializer
from .signer import HMACAlgorithm
from .signer import NoneAlgorithm
from .signer import Signer
from .timed import TimedSerializer
from .timed import TimestampSigner
from .url_safe import URLSafeSerializer
from .url_safe import URLSafeTimedSerializer

__version__ = "1.1.0"



import json
import sysactions
import time

from abc import ABCMeta
from xml.etree import cElementTree as eTree

from dateutil import tz
from msgbus import (
    TK_SYS_ACK,
    FM_PARAM,
    TK_HV_COMPLIST,
    MBEasyContext,
    TK_HV_GLOBALRESTART,
)
from persistence import Connection, Driver
from sysactions import (
    show_any_dialog,
    show_keyboard,
    get_model,
    get_pricelist,
    get_podtype,
    get_custom,
    get_current_order,
    show_messagebox,
    translate_message,
)
from systools import sys_log_error
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Union,
)

MODIFIER_PREFIXES = {}

class BaseRepository(object):
    __metaclass__ = ABCMeta

    def __init__(self, mbcontext):
        # type: (MBEasyContext) -> None
        self.mbcontext = mbcontext

    def execute_with_connection(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], int, str) -> Any
        conn = None
        try:
            conn = Driver().open(self.mbcontext,
                                 dbname=str(db_name) if db_name is not None else None,
                                 service_name=service_name)
            return method(conn)
        finally:
            if conn is not None:
                conn.close()

    def execute_with_transaction(self, method, db_name=None, service_name=None):
        # type: (Callable[[Connection], Any], int, str) -> Any
        def transaction_method(conn):
            # type: (Connection) -> Any

            conn.transaction_start()
            try:
                conn.query("BEGIN TRANSACTION")
                ret = method(conn)
                conn.query("COMMIT")
                return ret
            except:
                conn.query("ROLLBACK")
                raise
            finally:
                conn.transaction_end()

        return self.execute_with_connection(transaction_method, db_name=db_name, service_name=service_name)

    def execute_in_all_databases(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        ret = []
        for pos in pos_list:
            conn = None
            try:
                conn = Driver().open(self.mbcontext, dbname=str(pos))
                ret.append(method(conn))
            finally:
                if conn is not None:
                    conn.close()

        return ret

    def execute_in_all_databases_returning_flat_list(self, method, pos_list):
        # type: (Callable[[Connection], Any], List[int]) -> Any
        inner_ret = self.execute_in_all_databases(method, pos_list)

        ret = []
        for inner_list in inner_ret:
            ret.extend(inner_list)

        return ret


def show_message_dialog(pos_id, title="$INFORMATION", message="", mask="", assync=False):
    return show_any_dialog(pos_id, "messageDialog", message, title, "", "", "", 600000, mask, "", "", "", None, assync)


def show_message_options_dialog(pos_id, buttons, title="$INFORMATION", message="", timeout=600000):
    resp = show_any_dialog(
        pos_id, "messageOptionsDialog", message, title, "", buttons, "", timeout, "", "", "", "", None, False)

    if resp in (None, ''):
        return resp

    return int(resp)


def show_numpad_dialog(pos_id, title="", message="", mask="INTEGER", default_value=""):
    resp = show_keyboard(pos_id, message, title=title, mask=mask, numpad=True, defvalue=default_value)

    if resp in (None, '', 'NaN'):
        return resp

    mask = mask.lower()
    if mask == 'password' or mask == 'login':
        return str(resp)

    if mask != "currency" and mask != "number":
        return int(resp)

    return float(resp)


def show_filtered_list_box_dialog(pos_id, options, title="", message="", mask="", buttons="$OK|$CANCEL"):
    if type(options) == list:
        options = "|".join(options)
    resp = show_any_dialog(
        pos_id, "filteredListBox", message, title, options, buttons, "",  600000, mask, "", "", "", None, False)

    if resp in (None, ''):
        return resp

    return int(resp)


def show_check_list_box_dialog(pos_id, options, title="", message="", mask="", buttons="$OK|$CANCEL"):
    if type(options) == list:
        options = "||".join(options)
    resp = show_any_dialog(
        pos_id, "checkListBox", message, title, options, buttons, "",  600000, mask, "", "", "", None, False)

    if resp in (None, ''):
        return []

    resp = resp.split(';')
    return resp


def show_editable_list_box_dialog(pos_id, options="", title="", buttons="$OK|$CANCEL"):
    resp = show_any_dialog(
        pos_id, "multipleColumnListBox", "", title, options, buttons, "",  600000, "", "", "", "", None, False)

    return resp


def show_bordereau_dialog(pos_id):
    return show_any_dialog(
        pos_id, "bordereauDialog", '', "$BORDEREAU", "", "", "", 6000000, "mask", "", "", "", None, False)


def show_tab_customer_dialog(pos_id):
    return show_any_dialog(
        pos_id, "tabCustomerInfoDialog", '', "$CUSTOMER_INFO", "", "", "", 600000, "mask", "", "", "", None, False)


def show_confirm_dialog(pos_id, title=''):
    if not title:
        title = "$BORDEREAU"
    return show_any_dialog(
        pos_id, "bordereauDialog", 'confirmDialog', title, "", "", "", 3600000, "mask", "", "", "", None, False)


def show_qr_code(
    pos_id,             # type: str
    code_data,          # type: str
    buttons="$CANCEL",  # type: str
    title="",           # type: Optional[str]
    timeout=600000,     # type: Optional[int]
    info_text="",       # type: Optional[str]
):
    # type: (...) -> Optional[int]

    resp = show_any_dialog(
        pos_id,
        "qrcode",
        code_data,
        title,
        info_text,
        buttons,
        "",
        timeout,
        "",
        "",
        "",
        "",
        "",
        True
    )
    if resp in (None, ''):
        return 0

    return int(resp)


def show_listbox_with_disabled_options(
    pos_id,                         # type: str
    options,                        # type: List
    message='$SELECT_AN_OPTION',    # type: Optional[str]
    title='',                       # type: Optional[str]
    defvalue='',                    # type: Optional[str]
    buttons='$OK|$CANCEL',          # type: Optional[str]
    icon='',                        # type: Optional[str]
    timeout=60000,                  # type: Optional[int]
    mask=''                         # type: Optional[List]
):
    # type: (...) -> Optional[Dict]

    minvalue, maxvalue, contents, asynch = ('', '', None, False)
    info_text = '|'.join(options) if hasattr(options, '__iter__') else str(options)
    mask_text = '|'.join(mask) if hasattr(mask, '__iter__') else str(mask)
    data = show_any_dialog(
        posid=pos_id,
        dlgtype='listbox',
        message=message,
        title=title,
        infotext=info_text,
        buttons=buttons,
        icon=icon,
        timeout=timeout,
        mask=mask_text,
        defvalue=defvalue,
        minvalue=minvalue,
        maxvalue=maxvalue,
        contents=contents,
        asynch=asynch
    )
    if asynch:
        return listbox_response(
            selected_list_index=int(data),
        )
    elif data is None:
        return None
    split = data.split(',')
    buttons_quantity = len(buttons.split("|"))
    if int(split[0]) != buttons_quantity - 1 and split[1] not in mask:
        return listbox_response(
            selected_list_index=int(split[1]),
            selected_button_index=int(split[0]),
        )
    else:
        return None


def listbox_response(
    selected_list_index=None,       # type: Optional[int]
    selected_button_index=None,     # type: Optional[int]
):
    # type: (...) -> Dict

    listbox_response_dict = {
        "selected_list_index": selected_list_index,
        "selected_button_index": selected_button_index,
    }

    return listbox_response_dict


def show_device_manager_dialog(
    pos_id,                         # type: str
    info_text,                      # type: json
    message='',                     # type: Optional[str]
    title='$DEVICE_MANAGER',        # type: Optional[str]
    defvalue='',                    # type: Optional[str]
    buttons='$CANCEL|$UPDATE',      # type: Optional[str]
    icon='',                        # type: Optional[str]
    timeout=60000,                  # type: Optional[int]
):
    # type: (...) -> Optional[str]

    mask_text, minvalue, maxvalue, contents, asynch = ('', '', '', None, False)
    info_text = info_text.encode('UTF-8')
    data = show_any_dialog(
        posid=pos_id,
        dlgtype='deviceManager',
        message=message,
        title=title,
        infotext=info_text,
        buttons=buttons,
        icon=icon,
        timeout=timeout,
        mask=mask_text,
        defvalue=defvalue,
        minvalue=minvalue,
        maxvalue=maxvalue,
        contents=contents,
        asynch=asynch,
    )
    if data in (None, ''):
        return None

    return data


def show_loyalty_dialog(
    pos_id,        # type: str
    buttons,       # type: str
    message="",    # type: str
):
    # type: (...) -> Optional[Any[Union[int, str]]]

    resp = show_any_dialog(
        posid=pos_id,
        dlgtype='loyaltyDialog',
        message=message,
        title='$LOYALTY',
        infotext="",
        buttons=buttons,
        icon="",
        timeout=600000,
        mask="",
        defvalue="",
        minvalue="",
        maxvalue="",
        contents=None,
        asynch=False,
    )

    if resp in (None, ''):
        return resp

    return int(resp)


def components_is_running(mb_context, *params):
    msg_resp = mb_context.MB_EasySendMessage("HyperVisor", TK_HV_COMPLIST, FM_PARAM)
    if msg_resp.token != TK_SYS_ACK:
        sys_log_error("Unable to obtain the TK_HV_COMPLIST")
        return False

    component_list = [_.get('name').lower() for _ in eTree.XML(msg_resp.data).findall(".//component[@status='1']")]
    for component in params:
        if component.lower() not in component_list:
            return False

    return True


def ensure_iterable(value):
    if isinstance(value, basestring):
        return filter(None, [value])

    try:
        _ = (e for e in value)
        return filter(None, value)
    except TypeError:
        return filter(None, [value])


def ensure_list(value):
    if isinstance(value, basestring):
        return [value]

    if isinstance(value, list):
        return value

    try:
        _ = (e for e in value)
        return list(value)
    except TypeError:
        return [value]


def convert_to_dict(iterable):
    ret = {}
    for item in iterable:
        ret[item] = item
    return ret


def convert_from_localtime_to_utc(localtime):
    # type: (datetime) -> datetime
    from_zone = tz.tzlocal()
    to_zone = tz.tzutc()
    localtime = localtime.replace(tzinfo=from_zone)
    utc = localtime.astimezone(to_zone)

    return utc


def restart_system(
        mb_context,             # type: MBEasyContext
        event_type,             # type: str
        pos_id="",              # type: str
        menu_manager_user="",   # type: str
):
    # type: (...) -> MBMessage

    _notify_global_restart(event_type, mb_context, menu_manager_user, pos_id)

    show_messagebox(
        posid=pos_id,
        message="$PLEASE_WAIT_ALL_NODES_RESTART",
    )
    return mb_context.MB_SendMessage(
        destination=mb_context.hv_service,
        token=TK_HV_GLOBALRESTART,
    )


def _notify_global_restart(event_type, mb_context, menu_manager_user, pos_id):
    event_xml = '''<GLOBALRESTART posId="{}" userId="{}"/>'''
    data = event_xml.format(pos_id, menu_manager_user)
    mb_context.MB_EasyEvtSend(
        subject="GLOBAL_RESTART",
        type=event_type,
        xml=data,
        synchronous=True,
        timeout=-1,
    )

class ModelService(object):

    def __init__(
        self,
        pos_id,  # type: Any[Union[str, int]]
    ):
        # type: (...) -> None

        self.pos_id = pos_id

    def get_price_list_from_model(
        self,
        pod_type=None,    # type: Optional[str]
        sale_type=None,   # type: Optional[str]
    ):
        # type: (...) -> str

        model = get_model(posid=self.pos_id)
        if not pod_type:
            pod_type = get_podtype(model=model)

        price_list = get_pricelist(model=model)

        is_delivery = pod_type == "DL" or sale_type == "DELIVERY"
        if is_delivery and "DL" not in price_list:
            price_list = "DL." + price_list if price_list else "DL"

        special_catalog_price_list = self._get_price_list_from_special_catalog(model=model)
        if special_catalog_price_list:
            price_list = special_catalog_price_list + "." + price_list

        if "EI" not in price_list:
            price_list += ".EI"

        return price_list

    def _get_price_list_from_special_catalog(
        self,
        model,  # type: eTree
    ):
        # type: (...) -> str

        price_list = None
        special_catalog = json.loads(self.get_special_catalog(model=model))
        price_list = self._get_table_special_catalog(
            model=model,
            price_list=price_list,
            special_catalog=special_catalog,
        )
        if not price_list and "servicePeriod" in special_catalog.keys():
            price_list = special_catalog["servicePeriod"]

        return price_list

    def _get_table_special_catalog(
        self,
        model,            # type: eTree
        price_list,       # type: str
        special_catalog,  # type: dict
    ):
        # type: (...) -> Optional[str]

        current_order = model.find("CurrentOrder/Order")
        if current_order is None:
            return price_list

        pos_ot = sysactions.get_posot(model)
        order_id = current_order.get("orderId")
        order_picture = eTree.XML(pos_ot.orderPicture(orderid=order_id)).find("Order")
        if order_picture is None:
            return price_list

        order_state = order_picture.get("state")
        if order_state == "IN_PROGRESS" and self.is_table_service(order_picture=order_picture):
            order_properties = order_picture.find("CustomOrderProperties")
            table_id = order_properties.find("OrderProperty[@key='TABLE_ID']").get('value')
            if table_id is None:
                return price_list

            table = model.find("STORE_MODIFIED/Tables/Table[@id='" + table_id + "']")
            table_service_period = table.find("Properties").get("SERVICE_PERIOD_TABLE")
            if table_service_period is not None:
                price_list = json.loads(table_service_period).get("servicePeriod")
            else:
                sector_property = order_properties.find("OrderProperty[@key='SECTOR']")
                if special_catalog.get("sectors") is not None and sector_property is not None:
                    sector = sector_property.get('value')
                    if special_catalog["sectors"].get(sector) is not None:
                        price_list = special_catalog["sectors"][sector].get("servicePeriod")

        return price_list

    def get_special_catalog(
        self,
        model=None,  # type: Optional[eTree]
    ):
        # type: (...) -> str

        if model is None:
            model = get_model(posid=self.pos_id)

        special_catalog = get_custom(
            model=model,
            key='SERVICE_PERIOD',
        )
        if special_catalog in ['', None]:
            return '{}'

        return special_catalog

    @staticmethod
    def is_table_service(
        order_picture,  # type: eTree
    ):
        # type: (...) -> bool

        order_properties = order_picture.find("CustomOrderProperties")
        table_id = order_properties.find("OrderProperty[@key='TABLE_ID']") if order_properties is not None else None
        if table_id is not None:
            return True

        return False


def get_modifier_prefixes():
    # type: (...) -> Dict[str, str]

    global MODIFIER_PREFIXES

    if MODIFIER_PREFIXES:
        return MODIFIER_PREFIXES

    MODIFIER_PREFIXES = {
        "[OnSide]": "$ONSIDE",
        "[Extra]": "$EXTRA",
        "[Light]": "$LIGHT",
        "[Additional]": "$ADDITIONAL",
        "[Without]": "$WITHOUT",
    }

    store_prefixes = sysactions.get_storewide_config(
        parameter="Modifiers.Prefixes",
        defval=[],
        array=True,
    )
    for prefix in store_prefixes:
        values = prefix.split(";") if ";" in prefix else None
        if values:
            MODIFIER_PREFIXES[values[0]] = values[1]

    return MODIFIER_PREFIXES


def format_currency(
    model,  # type: eTree.XML
    value,  # type: float
):
    # type: (...) -> str

    currency_symbol = translate_message(
        model=model,
        message_key="L10N_CURRENCY_SYMBOL",
    )

    return "{0} {1:.2f}".format(
        currency_symbol,
        value
    )


def split_and_translate_message(
    model,      # type: str
    message,    # type: str
):
    # type: (...) -> str

    message = message[1:] if message.startswith("$") else message
    params = ""
    if "|" in message:
        message, params = message.split("|", 1)

    return translate_message(model, message, *(params.split("|")))



from ._BenefitApplierNotFound import BenefitApplierNotFound
from ._CouponAlreadyUsed import CouponAlreadyUsed
from ._CouponCustomError import CouponCustomError
from ._CouponInternalError import CouponInternalError
from ._CouponLocked import CouponLocked
from ._CouponNotFound import CouponNotFound
from ._TranslatedMessage import TranslatedMessage



from typing import List


class FiscalOrderResponse(object):
    def __init__(
        self,
        fiscal_coupon,          # type: str
        tef_customer_receipt,   # type: List[str]
    ):
        # type: (...) -> None

        self.fiscalCoupon = fiscal_coupon
        self.tefCustomerReceipt = tef_customer_receipt




from datetime import datetime
from typing import Union


class Order(object):
    def __init__(self, order_id, void_reason, operator_id, authorizer_id, order_datetime, total_gross):
        # type: (int, unicode, int, Union[int, None], datetime, float) -> None
        self.id = order_id
        self.reason = void_reason
        self.operator = operator_id
        self.authorizer = authorizer_id
        self.datetime = order_datetime
        self.total = total_gross



from typing import Dict


class AdditionalInfo(object):
    def __init__(
        self,
        additional_info,    # type: Dict
    ):
        # type: (...) -> None

        self.welcome_message = additional_info.get("welcomeMessage", None)
        self.end_sale_message = additional_info.get("endSaleMessage", None)
